//
//  ZRModel.m
//  AsyncDispatch
//
//  Created by ampatspell on 7/23/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "ZRModel.h"
#import "ZRModelOperation.h"
#import "ZRMethodImplementationType.h"
#import "ZRConfiguration.h"
#import "ZRModelConfiguration.h"
#import "ZRModelProperty.h"
#import "ZRRoute.h"
#import "ZRQuery.h"


static void *ZRModelMarkAsDirtyContext = (void *) @"ZRModelMarkAsDirtyContext"; 


@interface ZRModel ()
- (void)_startObservingPersistentProperties;
- (void)_stopObservingPersistentProperties;
@end

@implementation ZRModel
@synthesize state, key, error, disableDirtyRegistration;

+ (NSSet *)keyPathsForValuesAffectingBusy
{
  return [NSSet setWithObjects:@"state", nil];
}

+ (NSSet *)keyPathsForValuesAffectingDirty
{
  return [NSSet setWithObjects:@"state", nil];
}

+ (NSSet *)keyPathsForValuesAffectingCreated
{
  return [NSSet setWithObjects:@"state", nil];
}

+ (NSSet *)keyPathsForValuesAffectingDestroyed
{
  return [NSSet setWithObjects:@"state", nil];
}

#pragma mark -

+ (void)initialize
{
  ZRMethodImplementationType type = ZRClassMethodImplementationType(self, @selector(prepare));
  if(type == ZRMethodImplementationTypeImplement || type == ZRMethodImplementationTypeOverride) {
    [self prepare];
  }
}

+ (void)prepare
{
}

+ (ZRModelConfiguration *)modelConfiguration
{
  return [[ZRConfiguration sharedConfiguration] modelConfigurationForClass:self];
}

+ (void)hasProperty:(NSString *)propertyName withServerName:(NSString *)serverName
{
  ZRModelProperty *property = [[ZRModelProperty alloc] initWithName:propertyName 
                                                         serverName:serverName 
                                                           forClass:[self class]];
  [[self modelConfiguration] addProperty:property];
  [property release];
}

+ (void)hasProperty:(NSString *)propertyName
{
  [self hasProperty:propertyName withServerName:propertyName];
}

+ (void)supports:(ZRRouteKind)kind withPath:(NSString *)path
{
  ZRRoute *route = [[ZRRoute alloc] initWithKind:kind path:path];
  [[self modelConfiguration] addRoute:route];
  [route release];
}

+ (void)query:(ZRQueryKind)kind withPath:(NSString *)path arrayClass:(Class)arrayClass
{
  ZRQuery *query = [[ZRQuery alloc] initWithKind:kind path:path arrayClass:arrayClass];
  [[self modelConfiguration] addQuery:query];
  [query release];
}

#pragma mark -

- (id)init
{
  return [self initWithKey:nil state:0 delegate:nil];
}

- (id)initWithKey:(NSString *)aKey state:(ZRModelState)aState delegate:(id<ZRModelDelegate>)delegate
{
  NSParameterAssert(delegate);
  if(self = [super init]) {
    key           = [aKey copy];
    state         = aState;
    modelDelegate = delegate;
    [self _startObservingPersistentProperties];
  }
  return self;
}

- (void)dealloc
{
  [self _stopObservingPersistentProperties];
  modelDelegate = nil;
  self.key      = nil;
  self.error    = nil;
  [super dealloc];
}

#pragma mark -

- (void)_startObservingPersistentProperties
{
  for(ZRModelProperty *property in [[self class] modelConfiguration].properties) {
    [self addObserver:self forKeyPath:property.name options:NSKeyValueObservingOptionNew context:&ZRModelMarkAsDirtyContext];
  }
}

- (void)_stopObservingPersistentProperties
{
  for(ZRModelProperty *property in [[self class] modelConfiguration].properties) {
    [self removeObserver:self forKeyPath:property.name];
  }
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
  if(context == &ZRModelMarkAsDirtyContext) {
    [self markAsDirty];
  } else {
    [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
  }
}

- (void)addObserver:(NSObject *)observer forKeyPath:(NSString *)keyPath options:(NSKeyValueObservingOptions)options context:(void *)context
{
  [super addObserver:observer forKeyPath:keyPath options:options context:context];
  if(context != &ZRModelMarkAsDirtyContext) {
    [self fetch];    
  }
}

#pragma mark -

- (BOOL)hasState:(int)flag
{
  return ZRModelStateHas(state, flag);
}

- (BOOL)isBusy
{
  return [self hasState:ZRModelStateBusy];
}

- (BOOL)isDirty
{
  return [self hasState:ZRModelStateDirty];
}

- (BOOL)isDestroyed
{
  return [self hasState:ZRModelStateDestroyed];
}

- (BOOL)isCreated
{
  return [self hasState:ZRModelStateCreated];
}

- (void)markAsDirty
{
  if(!disableDirtyRegistration) {
    self.state = ZRModelStateAdd(state, ZRModelStateDirty);    
  }
}

#pragma mark -

- (NSDictionary *)_keyOnlyServerParameters
{
  id value = [self valueForKey:@"key"];
  if(value) {
    ZRModelProperty *property = [[[self class] modelConfiguration] propertyForName:@"key"];
    NSString *serverName;
    if(property) {
      serverName = property.serverName;
    } else {
      serverName = @"key";
    }
    return [NSDictionary dictionaryWithObject:value forKey:serverName];
  }
  return nil;
}

- (NSDictionary *)_serverParameters
{
  NSMutableDictionary *dictionary = [NSMutableDictionary dictionary];
  ZRModelConfiguration *current = [[self class] modelConfiguration];
  while(current) {
    for(ZRModelProperty *property in current.properties) {
      id value = [self valueForKey:property.name];
      if(value) {
        [dictionary setObject:value forKey:property.serverName];
      }
    }
    current = [current superclassConfiguration];
  }
  return dictionary;
}

#pragma mark -

- (ZRRouteKind)_routeKindForCurrentBusyState
{
  if([self hasState:ZRModelStateLoading]) {
    return ZRRouteKindLoad;
  } else if([self hasState:ZRModelStateCreating]) {
    return ZRRouteKindCreate;
  } else if([self hasState:ZRModelStateUpdating]) {
    return ZRRouteKindUpdate;
  } else if([self hasState:ZRModelStateDestroying]) {
    return ZRRouteKindDestroy;
  } 
  return nil;
}

- (void)_performOperation:(ZRRouteKind)kind 
               parameters:(NSDictionary *)parameters 
                   target:(id)target 
                  success:(SEL)success 
                   failed:(SEL)failed
{
  ZRRoute *route = [[[self class] modelConfiguration] routeForKind:kind];
  if(!route) {
    return;
  }
  
  self.state = route.busyState;
  
  ZRModelOperation *operation = [[ZRModelOperation alloc] init];
  operation.model      = self;
  operation.action     = route.path;
  operation.parameters = parameters;
  operation.target     = target;
  operation.success    = success;
  operation.failed     = failed;
  [modelDelegate performOperation:operation];
  [operation release];
}

- (void)reloadWithTarget:(id)target success:(SEL)success failed:(SEL)failed
{
  if(self.busy || !self.key || self.destroyed) {
    return;
  }

  NSDictionary *parameters = [self _serverParameters];
  
  [self _performOperation:ZRRouteKindLoad 
               parameters:parameters 
                   target:target 
                  success:success 
                   failed:failed];
}

- (void)reload
{
  [self reloadWithTarget:nil success:nil failed:nil];
}

- (void)fetchWithTarget:(id)target success:(SEL)success failed:(SEL)failed
{
  if([self hasState:ZRModelStateBlank]) {
    [self reloadWithTarget:target success:success failed:failed];
  }
}

- (void)fetch
{
  [self fetchWithTarget:nil success:nil failed:nil];
}

- (void)commitWithTarget:(id)target success:(SEL)success failed:(SEL)failed
{
  if(self.busy || self.destroyed) {
    return;
  }

  if(!([self hasState:ZRModelStateDirty] || [self hasState:ZRModelStateError])) {
    return;
  }
  
  ZRRouteKind kind;
  if([self hasState:ZRModelStateNew]) {
    kind = ZRRouteKindCreate;
  } else if([self hasState:ZRModelStateCreated]) {
    kind = ZRRouteKindUpdate;
  } else {
    return;
  }
  
  NSDictionary *parameters = [self _serverParameters];
  
  [self _performOperation:kind 
               parameters:parameters 
                   target:target 
                  success:success 
                   failed:failed];
}

- (void)commit
{
  [self commitWithTarget:nil success:nil failed:nil];
}

- (void)destroyWithTarget:(id)target success:(SEL)success failed:(SEL)failed
{
  if(self.busy || self.destroyed || !self.created) {
    return;
  }
  
  [self _performOperation:ZRRouteKindDestroy 
               parameters:[self _keyOnlyServerParameters] 
                   target:target 
                  success:success 
                   failed:failed];
}

- (void)destroy
{
  [self destroyWithTarget:nil success:nil failed:nil];
}

- (void)_setModelPropertiesFromDictionary:(NSDictionary *)dictionary usingConfiguration:(ZRModelConfiguration *)configuration
{
  if(dictionary) {
    ZRModelConfiguration *current = configuration;
    while(current) {
      for(ZRModelProperty *property in current.properties) {
        id value = [dictionary objectForKey:property.serverName];
        if(value) {
          [self setValue:value forKey:property.name];
        }        
      }
      current = [current superclassConfiguration];
    }
  }  
}

- (void)setModelPropertiesFromDictionary:(NSDictionary *)dictionary
{
  [self _setModelPropertiesFromDictionary:dictionary usingConfiguration:[[self class] modelConfiguration]];
}

- (void)modelOperation:(ZRModelOperation *)operation didSucceedWithResult:(id)result
{
  ZRRouteKind kind = [self _routeKindForCurrentBusyState];  
  ZRModelConfiguration *configuration = [[self class] modelConfiguration];
  ZRRoute *route = [configuration routeForKind:kind];
  if(!route) {
    return;
  }
  
  NSError *nextError = nil;
  if(![self isResultSuccessful:result error:&nextError]) {
    self.error = nextError;
    self.state = route.failedState;
    [self performSelector:route.failedSelector];
    [operation performTargetSelector:NO];
    return;
  }
  
  NSDictionary *data = [self dataDictionaryFromResult:result];
  [self _setModelPropertiesFromDictionary:data usingConfiguration:configuration];
  
  self.error = nil;
  self.state = route.successfulState;
  [self performSelector:route.successfulSelector];
  [operation performTargetSelector:YES];
}

- (void)modelOperation:(ZRModelOperation *)operation didFailWithError:(NSError *)nextError
{
  if([modelDelegate retryOperation:operation withError:nextError]) {
    self.state = ZRModelStateAdd(state, ZRModelStateRetrying);
    return;
  }
  
  ZRRouteKind kind = [self _routeKindForCurrentBusyState];

  ZRRoute *route = [[[self class] modelConfiguration] routeForKind:kind];
  if(!route) {
    return;
  }
  
  self.error = nextError;
  self.state = route.failedState;
  [self performSelector:route.failedSelector];
  [operation performTargetSelector:NO];
}

#pragma mark -

- (NSString *)description
{
  return [NSString stringWithFormat:@"<%@ %p: (%@) %@>", [self className], self, 
          NSStringFromZRModelState(self.state), self.key];
}

@end


@implementation ZRModel (ZRModelBlankImplemetations)
- (void)didLoad {}
- (void)didCreate {}
- (void)didUpdate {}
- (void)didDestroy {}
- (void)loadDidFail {}
- (void)createDidFail {}
- (void)updateDidFail {}
- (void)destroyDidFail {}
@end
