//
//  ZRModelArray.m
//  AsyncDispatch
//
//  Created by ampatspell on 7/25/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "ZRModelArray.h"
#import "ZRModelArrayOperation.h"
#import "ZRQuery.h"


@interface ZRModelArray ()
@property(nonatomic, readwrite) ZRModelState state;
@end

@implementation ZRModelArray
@synthesize state, error;

- (id)init
{
  return [self initWithModelClass:nil query:nil parameters:nil state:0 delegate:nil];
}

- (id)initWithModelClass:(Class)aModelClass
                   query:(ZRQuery *)aQuery 
              parameters:(NSDictionary *)aParameters
                   state:(ZRModelState)aState 
                delegate:(id<ZRModelArrayDelegate>)aDelegate;
{
  NSParameterAssert(aDelegate != nil);
  if(self = [super init]) {
    modelClass = aModelClass;
    query = [aQuery retain];
    queryParameters = [aParameters copy];
    state = aState;
    arrayDelegate = aDelegate;
    models = [[NSMutableArray alloc] init];    
  }
  return self;
}

- (void)dealloc
{
  arrayDelegate = nil;
  [query release], query = nil;
  [queryParameters release], queryParameters = nil;
  [models release], models = nil;
  [error release], error = nil;
  modelClass = nil;
  [super dealloc];  
}

#pragma mark -

- (BOOL)isBusy
{
  return [self hasState:ZRModelStateBusy];
}

- (BOOL)hasState:(NSUInteger)flag
{
  return ZRModelStateHas(self.state, flag);
}

#pragma mark -

- (NSUInteger)countByEnumeratingWithState:(NSFastEnumerationState *)enumerationState objects:(id *)stackbuf count:(NSUInteger)len
{
  return [models countByEnumeratingWithState:enumerationState objects:stackbuf count:len];
}

- (NSUInteger)countOfModels
{
  return [models count];
}

- (id)objectInModelsAtIndex:(NSUInteger)index
{
  return [models objectAtIndex:index];
}

- (NSArray *)modelsAtIndexes:(NSIndexSet *)indexes
{
  return [models objectsAtIndexes:indexes];
}

- (void)insertObject:(id)object inModelsAtIndex:(NSUInteger)index
{
  [models insertObject:object atIndex:index];
}

- (void)removeObjectFromModelsAtIndex:(NSUInteger)index
{
  [models removeObjectAtIndex:index];
}

- (void)replaceObjectsAtIndexes:(NSIndexSet *)indexes withModels:(NSArray *)objects
{
  [models replaceObjectsAtIndexes:indexes withObjects:objects];
}

- (void)insertModels:(NSArray *)array atIndexes:(NSIndexSet *)indexes
{
  [models insertObjects:array atIndexes:indexes];
}

- (void)removeModelsAtIndexes:(NSIndexSet *)indexes
{
  [models removeObjectsAtIndexes:indexes];
}

#pragma mark -

- (void)addModel:(ZRModel *)model
{
  NSIndexSet *indexes = [NSIndexSet indexSetWithIndex:[models count]];
  [self willChange:NSKeyValueChangeInsertion valuesAtIndexes:indexes forKey:@"models"];
  [models addObject:model];
  [self didChange:NSKeyValueChangeInsertion valuesAtIndexes:indexes forKey:@"models"];
}

- (void)removeModel:(ZRModel *)model
{
  NSUInteger index = [models indexOfObject:model];
  if(index != NSNotFound) {
    NSIndexSet *indexes = [NSIndexSet indexSetWithIndex:index];
    [self willChange:NSKeyValueChangeRemoval valuesAtIndexes:indexes forKey:@"models"];
    [models removeObject:model];    
    [self didChange:NSKeyValueChangeRemoval valuesAtIndexes:indexes forKey:@"models"];
  }
}

- (void)removeAllModels
{
  NSUInteger count = [self countOfModels];
  if(count > 0) {
    NSIndexSet *indexes = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, count)];
    [self willChange:NSKeyValueChangeRemoval valuesAtIndexes:indexes forKey:@"models"];
    [models removeAllObjects];
    [self didChange:NSKeyValueChangeRemoval valuesAtIndexes:indexes forKey:@"models"];
  }
}

#pragma mark -

- (NSDictionary *)_operationParameters
{
  NSMutableDictionary *params = [NSMutableDictionary dictionary];
  
  if(queryParameters) {
    [params addEntriesFromDictionary:queryParameters];
  }
  
  NSDictionary *parameters = [modelClass parametersForModelArray:self];
  
  if(parameters) {
    [params addEntriesFromDictionary:parameters];
  }
  
  return params;
}

- (void)_performWithTarget:(id)target success:(SEL)success failed:(SEL)failed clear:(BOOL)clear
{
  if(self.busy) {
    return;
  }
  
  if(clear) {
    [self willReload];
  }
  
  ZRModelArrayOperation *operation = [[ZRModelArrayOperation alloc] init];
  
  operation.array      = self;
  operation.clear      = clear;
  operation.action     = query.path;
  operation.parameters = [self _operationParameters];
  operation.target     = target;
  operation.success    = success;
  operation.failed     = failed;
  
  self.state = ZRModelStateBusy | ZRModelStateLoading;
  
  [arrayDelegate performOperation:operation];
  [operation release];
}

- (void)reloadWithTarget:(id)target success:(SEL)success failed:(SEL)failed
{
  [self _performWithTarget:target success:success failed:failed clear:YES];
}

- (void)reload
{
  [self reloadWithTarget:nil success:nil failed:nil];
}

- (void)fetchWithTarget:(id)target success:(SEL)success failed:(SEL)failed
{
  if([self hasState:ZRModelStateBlank]) {
    [self _performWithTarget:target success:success failed:failed clear:NO];
  }
}

- (void)fetch
{
  [self fetchWithTarget:nil success:nil failed:nil];
}

#pragma mark -

- (void)modelArrayOperation:(ZRModelArrayOperation *)operation didSucceedWithResult:(id)result
{
  NSError *nextError = nil;
  if(![modelClass isResultSuccessful:result error:&nextError forModelArray:self]) {
    self.error = nextError;
    if([self countOfModels] == 0) {
      self.state = ZRModelStateIdle | ZRModelStateError;      
    } else {
      self.state = ZRModelStateBlank | ZRModelStateIdle | ZRModelStateError;
    }
    [operation performTargetSelector:NO];
    return;
  }
  
  if(operation.clear) {
    [self removeAllModels];
  }
  
  NSArray *data = [modelClass dataArrayFromResult:result];
  NSMutableArray *newModels = [[NSMutableArray alloc] init];
  for(NSDictionary *dictionary in data) {
    id model = [arrayDelegate createdModelForClass:modelClass fromDictionary:dictionary];
    [newModels addObject:model];
  }
  
  NSIndexSet *indexes = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange([self countOfModels], [newModels count])];
  [self insertModels:newModels atIndexes:indexes];
    
  BOOL finished = [self isLoadingFinishedForOperation:operation withResultModels:newModels];
  [newModels release];
  
  if(finished) {
    self.error = nil;
    self.state = ZRModelStateIdle;
    [operation performTargetSelector:YES];
  } else {
    operation.parameters = [self _operationParameters];
    operation.clear = NO;
    [arrayDelegate performOperation:operation];
  }
}

- (void)modelArrayOperation:(ZRModelArrayOperation *)operation didFailWithError:(NSError *)nextError
{
  if([arrayDelegate retryOperation:operation withError:nextError]) {
    self.state = ZRModelStateAdd(state, ZRModelStateRetrying);
    return;
  }
  
  self.state = ZRModelStateBlank | ZRModelStateError;
  self.error = nextError;
  
  [operation performTargetSelector:NO];
}

#pragma mark -

- (id)modelForClass:(Class)modelClass fromDictionary:(NSDictionary *)dictionary
{
  return nil;
}

@end

