//
//  ZRRPCOperation.m
//  ZeroEdit
//
//  Created by ampatspell on 6/4/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "ZRRPCOperation.h"
#import "JSON.h"

NSString *const ZRRPCOperationErrorDomain = @"AMRPCOperationErrorDomain";

NSInteger const ZRRPCOperationConnectionFailedErrorCode = 10;
NSInteger const ZRRPCOperationInvalidJSONErrorCode = 11;
NSInteger const ZRRPCOperationResponseStatusErrorCode = 12;


@interface ZRRPCOperation ()
@property(nonatomic, readwrite, getter=isExecuting) BOOL executing;
@property(nonatomic, readwrite, getter=isFinished) BOOL finished;
@property(nonatomic, readwrite, retain) NSURLConnection *connection;
@property(nonatomic, readwrite) NSInteger status;
@property(nonatomic, readwrite, retain) NSMutableData *body;
- (void)sendRequest;

- (void)requestDidSucceed;
- (void)requestDidFailWithErrorCode:(NSInteger)errorCode description:(NSString *)description;
- (NSError *)errorWithCode:(NSInteger)code description:(NSString *)description;
- (void)requestDidFailWithError:(NSError *)error;
- (void)requestDidFinish;

- (void)RPCDidSucceedWithJSON:(id)json;
- (void)RPCDidFailWithError:(NSError *)error;
@end

@implementation ZRRPCOperation

@synthesize url = url_;
@synthesize parameters = parameters_;
@synthesize delegate = delegate_;
@synthesize connection = connection_;
@synthesize status = status_;
@synthesize body = body_;
@synthesize executing = executing_;
@synthesize finished = finished_;
@synthesize userInfo = userInfo_;

- (id)init
{
  if(self = [super init]) {
    self.executing = NO;
    self.finished = NO;
  }
  return self;
}

- (void)dealloc
{
  self.url = nil;
  self.parameters = nil;
  self.delegate = nil;
  self.userInfo = nil;
  [super dealloc];
}

#pragma mark -

- (BOOL)isConcurrent
{
  return YES;
}

- (void)start
{
  self.executing = YES;
  self.finished  = NO;
  [self sendRequest];
}

#pragma mark -

- (void)sendRequest
{
  NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:self.url];
  [request setCachePolicy:NSURLCacheStorageNotAllowed];
  [request setHTTPMethod:@"POST"];
  if(self.parameters) {
    [request setHTTPBody:[[self.parameters JSONRepresentation] dataUsingEncoding:NSUTF8StringEncoding]];
  }
  self.connection = [NSURLConnection connectionWithRequest:request delegate:self];
}

#pragma mark -

- (void)requestDidSucceed
{
  NSString *body = [[NSString alloc] initWithData:self.body encoding:NSUTF8StringEncoding];
  id json = [body JSONValue];
  if(json) {
    [self RPCDidSucceedWithJSON:json];
  } else {
    [self RPCDidFailWithError:[self errorWithCode:ZRRPCOperationInvalidJSONErrorCode 
                                      description:@"Server returned invalid JSON string"]];
  }
  [self requestDidFinish];
  [body release];
}

- (void)requestDidFailWithErrorCode:(NSInteger)errorCode description:(NSString *)description
{
  [self requestDidFailWithError:[self errorWithCode:errorCode description:description]];
}

- (NSError *)errorWithCode:(NSInteger)code description:(NSString *)description
{
  NSDictionary *userInfo = [NSDictionary dictionaryWithObject:description forKey:NSLocalizedDescriptionKey];
  return [NSError errorWithDomain:ZRRPCOperationErrorDomain code:code userInfo:userInfo];
}

- (void)requestDidFailWithError:(NSError *)error
{
  [self RPCDidFailWithError:error];
  [self requestDidFinish];   
}

- (void)requestDidFinish
{
  [self willChangeValueForKey:@"isFinished"];
  [self willChangeValueForKey:@"isExecuting"];
  executing_ = NO;
  finished_ = YES;
  [self didChangeValueForKey:@"isExecuting"];
  [self didChangeValueForKey:@"isFinished"];
  
  self.connection = nil;
  self.body = nil;
  self.userInfo = nil;
}

#pragma mark -

- (void)RPCDidSucceedWithJSON:(id)json
{
  if([self isCancelled]) {
    return;
  }
  
  [self.delegate RPCOperation:self didSucceed:json];
}

- (void)RPCDidFailWithError:(NSError *)error
{
  if([self isCancelled]) {
    return;
  }
  
  [self.delegate RPCOperation:self didFailWithError:error];
}

#pragma mark -

- (NSURLRequest *)connection:(NSURLConnection *)connection 
             willSendRequest:(NSURLRequest *)request 
            redirectResponse:(NSURLResponse *)response
{
  return request;
}

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
  if([response isKindOfClass:[NSHTTPURLResponse class]]) {
    NSHTTPURLResponse *httpResponse = (NSHTTPURLResponse *)response;
    self.status = [httpResponse statusCode];
    self.body = [[[NSMutableData alloc] init] autorelease];
  }
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
  [self.body appendData:data];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
  if(self.status == 200) {
    [self requestDidSucceed];
  } else {
    [self requestDidFailWithErrorCode:ZRRPCOperationResponseStatusErrorCode description:@"Server returned non 200 status code"];
  }
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
  [self requestDidFailWithErrorCode:ZRRPCOperationConnectionFailedErrorCode description:@"Cannot connect to server"];
}

@end
